/*
 * (C) Copyright IBM Corp. 2003
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   Konrad Rzeszutek <konradr@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.15.2.1 2003/02/27 05:59:41 rstory Exp $
 *
 * $Id: saHpiCtrlTable.c,v 1.18 2004/02/17 20:28:17 konradrzeszutek Exp $
 *
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <saHpiCtrlTable.h>
#include <hpiSubagent.h>
#include <SaHpi.h>
#include <netinet/in.h>


static netsnmp_handler_registration *my_handler = NULL;
static netsnmp_table_array_callbacks cb;

static oid saHpiCtrlTable_oid[] = { saHpiCtrlTable_TABLE_OID };
static size_t saHpiCtrlTable_oid_len = OID_LENGTH (saHpiCtrlTable_oid);

static oid saHpiCtrlCount_oid[] = { hpiResources_OID, 3, 0 };


static u_long control_count = 0;


static int
saHpiCtrlTable_modify_context (SaHpiEntryIdT rdr_id,
			       SaHpiCtrlRecT * entry,
			       SaHpiCtrlStateT * ctrl_state,
			       SaHpiRptEntryT * rpt_entry,
			       oid * rdr_entry, size_t rdr_entry_oid_len,
			       saHpiCtrlTable_context * ctx);



int
populate_control (SaHpiEntryIdT rdr_id,
		  SaHpiCtrlRecT * ctrl,
		  SaHpiRptEntryT * rpt_entry,
		  oid * rdr_entry_oid, size_t rdr_entry_oid_len,
		  oid * ctrl_oid, size_t * ctrl_oid_len)
{

  int rc = AGENT_ERR_NOERROR;
  oid index_oid[CTRL_INDEX_NR];
  oid column[2];
  SaHpiSessionIdT session_id;
#ifdef BUG_873961
  int i = 0;
  netsnmp_void_array *array;
  saHpiCtrlTable_context *ctx;
#endif
  netsnmp_index ctrl_index;
  saHpiCtrlTable_context *ctrl_context;
  SaHpiCtrlStateT ctrl_state;

  DEBUGMSGTL ((AGENT, "\n\t--- populate_ctrl: Entry.\n"));

  if (ctrl)
    {

      ctrl_index.len = CTRL_INDEX_NR;
      // Look at the MIB to find out what the indexs are
      index_oid[0] = rpt_entry->DomainId;
      index_oid[1] = rpt_entry->ResourceId;
      index_oid[2] = ctrl->Num;
      // Possible more indexes?
      ctrl_index.oids = (oid *) & index_oid;
      // We are re-populating. Check for existing entries
      ctrl_context = NULL;
      ctrl_context = CONTAINER_FIND (cb.container, &ctrl_index);
      // If we don't find it - we create it.
#ifdef BUG_873961      
      if (!ctrl_context)
	{
		// Bug # 873961. We use the 'rdr_id' to check to see if 
		// it is the context. To do so, we have to search for it first.
		ctrl_index.len = 1;
		// re-using the index_oid. 
		array = CONTAINER_GET_SUBSET (cb.container, &ctrl_index);
		if (array != NULL) 
		{
			if (array->size > 0) 
			{
				for (i = 0; i < array->size; i++) 
				{
					ctx = array->array[i];
					if (ctx->rdr_id == rdr_id) 
					{
						// Found the duplicate entry!
						ctrl_context = ctx;
						index_oid[1] = ctx->resource_id;
						index_oid[2] = ctx->saHpiCtrlNum;
						DEBUGMSGTL((AGENT,"duplicate control entry %d, %d, %d [rdr: %d] found.\n",
									rpt_entry->DomainId,
									rpt_entry->ResourceId,
									ctrl->Num,
									rdr_id));
						break;
					}
				}
			}
			free(array->array);
			free (array); array = NULL;
		}
		// restoree it to its previous glory.
		ctrl_index.len =  CTRL_INDEX_NR;
        } 
#endif
      if (!ctrl_context)
	{
	  // New entry. Add it
	  ctrl_context = saHpiCtrlTable_create_row (&ctrl_index);
	}

      if (!ctrl_context)
	{
	  snmp_log (LOG_ERR, "Not enough memory for a control row!");
	  return AGENT_ERR_INTERNAL_ERROR;
	}
      // Generate our full OID
      column[0] = 1;		// SaHpiCtrlTable 1
      column[1] = COLUMN_SAHPICTRLNUM;

      build_full_oid (saHpiCtrlTable_oid, saHpiCtrlTable_oid_len, column, 2,	// length of column is 2.
		      &ctrl_index, ctrl_oid, MAX_OID_LEN, ctrl_oid_len);

      // Get the Control State
      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}
      rc = saHpiControlStateGet (session_id,
				 rpt_entry->ResourceId,
				 ctrl->Num, &ctrl_state);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiControlStateGet failed with return code: %s.\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiControlStateGet failed with rc: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}

      // By this stage, ctrl_context surely has something in it.
      // '*_modify_context' does a checksum check to see if 
      // the record needs to be altered, and if so populates with
      // information from RDR and the OIDs passed.
      if (saHpiCtrlTable_modify_context (rdr_id,ctrl,
					 &ctrl_state,
					 rpt_entry,
					 rdr_entry_oid, rdr_entry_oid_len,
					 ctrl_context) == AGENT_NEW_ENTRY)
	{
	  CONTAINER_INSERT (cb.container, ctrl_context);
	  control_count = CONTAINER_SIZE (cb.container);
	}
      rc = AGENT_ERR_NOERROR;
    }
  else
    rc = AGENT_ERR_OPERATION;

  DEBUGMSGTL ((AGENT, "\n\t--- populate_ctrl. Exit\n"));
  return rc;
}

int
delete_ctrl_row (SaHpiDomainIdT domain_id,
		 SaHpiResourceIdT resource_id, SaHpiCtrlNumT num)
{

  saHpiCtrlTable_context *ctx;
  oid ctrl_oid[CTRL_INDEX_NR];
  netsnmp_index ctrl_index;
  int rc = AGENT_ERR_NOT_FOUND;

  DEBUGMSGTL ((AGENT, "delete_ctrl_row (%d, %d, %d). Entry\n",
	       domain_id, resource_id, num));
// Look at the MIB to find out what the indexs are
  ctrl_oid[0] = domain_id;
  ctrl_oid[1] = resource_id;
  ctrl_oid[2] = num;


  // Possible more indexs?
  ctrl_index.oids = (oid *) & ctrl_oid;
  ctrl_index.len = CTRL_INDEX_NR;

  ctx = CONTAINER_FIND (cb.container, &ctrl_index);

  if (ctx)
    {
      CONTAINER_REMOVE (cb.container, ctx);
      saHpiCtrlTable_delete_row (ctx);
      control_count = CONTAINER_SIZE (cb.container);
      rc = AGENT_ERR_NOERROR;
    }
  DEBUGMSGTL ((AGENT, "delete_ctrl_row. Exit (rc: %d)\n", rc));
  return rc;
}

int
read_textline (saHpiCtrlTable_context * ctx)
{

  SaHpiSessionIdT session_id;
  SaHpiCtrlStateT state;

  SaErrorT rc;

  if (ctx)
    {

      // Get the seesion_id
      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}
      memset (&state, 0x00, sizeof (SaHpiCtrlStateT));
      state.Type = ctx->saHpiCtrlType - 1;
      state.StateUnion.Text.Line = ctx->saHpiCtrlState[0];

      // This will read the requested line.
      rc = saHpiControlStateGet (session_id,
				 ctx->resource_id, ctx->saHpiCtrlNum, &state);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiControlStateGet failed with return code: %s\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to 'saHpiControlStateGet' failed. return code: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}
      // Copy the data in the columns.
      ctx->saHpiCtrlTextType = state.StateUnion.Text.Text.DataType+1;
      ctx->saHpiCtrlTextLanguage = state.StateUnion.Text.Text.Language+1;
      ctx->saHpiCtrlText_len =
	(state.StateUnion.Text.Text.DataLength <
	 SAHPI_TEXT_MAX) ? state.StateUnion.Text.Text.
	DataLength : SAHPI_TEXT_MAX;

      memcpy (ctx->saHpiCtrlText,
	      &state.StateUnion.Text.Text.Data, ctx->saHpiCtrlText_len);

      return AGENT_ERR_NOERROR;

    }
  return AGENT_ERR_NULL_DATA;
}

int
set_ctrl_state (saHpiCtrlTable_context * ctx)
{

  SaHpiSessionIdT session_id;
  SaHpiCtrlStateT state;

  SaErrorT rc;

  if (ctx)
    {

      // Get the seesion_id
      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}

      memset (&state, 0x00, sizeof (SaHpiCtrlStateT));
      rc = saHpiControlStateGet (session_id,
				 ctx->resource_id, ctx->saHpiCtrlNum, &state);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiControlStateGet failed with return code: %s\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiControlStateGet failed with return code: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}
      // To be compliant with the HPI.
      state.Type = ctx->saHpiCtrlType - 1;

      switch (state.Type)
	{
	case SAHPI_CTRL_TYPE_DIGITAL:
	  memcpy (&state.StateUnion.Digital,
		  ctx->saHpiCtrlState, sizeof (SaHpiCtrlStateDigitalT));
	  state.StateUnion.Digital = ntohl (state.StateUnion.Digital)-1;
	  break;
	case SAHPI_CTRL_TYPE_DISCRETE:
	  memcpy (&state.StateUnion.Discrete,
		  ctx->saHpiCtrlState, sizeof (SaHpiCtrlStateDiscreteT));
	  state.StateUnion.Discrete = ntohl (state.StateUnion.Discrete);
	  break;
	case SAHPI_CTRL_TYPE_ANALOG:
	  // This is Int32
	  memcpy (&state.StateUnion.Analog,
		  ctx->saHpiCtrlState, sizeof (SaHpiCtrlStateAnalogT));
	  state.StateUnion.Analog = ntohl (state.StateUnion.Analog);
	  break;
	case SAHPI_CTRL_TYPE_STREAM:
	  memcpy (&state.StateUnion.Stream,
		  ctx->saHpiCtrlState, sizeof (SaHpiCtrlStateStreamT));
	  state.StateUnion.Stream.StreamLength =
	    ntohl (state.StateUnion.Stream.StreamLength);
	  break;
	case SAHPI_CTRL_TYPE_TEXT:

	  state.StateUnion.Text.Line = ctx->saHpiCtrlState[0];
	  state.StateUnion.Text.Text.DataType = ctx->saHpiCtrlTextType - 1;
	  state.StateUnion.Text.Text.Language = ctx->saHpiCtrlTextLanguage - 1;
	  state.StateUnion.Text.Text.DataLength = ctx->saHpiCtrlText_len;
	  // reserve2 made sure that the length of data is less (or equal)
	  // than Text.Data can hold
	  memcpy (&state.StateUnion.Text.Text.Data,
		  ctx->saHpiCtrlText, ctx->saHpiCtrlText_len);

	  break;
	case SAHPI_CTRL_TYPE_OEM:
	  memcpy (&state.StateUnion.Oem.MId,
		  ctx->saHpiCtrlText, ctx->saHpiCtrlText_len);

	  state.StateUnion.Oem.MId = ntohl (state.StateUnion.Oem.MId);
	  state.StateUnion.Oem.BodyLength = ctx->saHpiCtrlText_len;
	  memcpy (&state.StateUnion.Oem.Body,
		  ctx->saHpiCtrlText, ctx->saHpiCtrlText_len);
	  break;
	}

      rc = saHpiControlStateSet (session_id,
				 ctx->resource_id, ctx->saHpiCtrlNum, &state);

      if (SA_OK != rc)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiControlStateSet failed with return code: %s\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiControlStateSet failed with return code %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}
      return AGENT_ERR_NOERROR;
    }
  return AGENT_ERR_NULL_DATA;
}

int
saHpiCtrlTable_modify_context (SaHpiEntryIdT rdr_id, SaHpiCtrlRecT * entry,
			       SaHpiCtrlStateT * state,
			       SaHpiRptEntryT * rpt_entry,
			       oid * rdr_entry, size_t rdr_entry_oid_len,
			       saHpiCtrlTable_context * ctx)
{

  long hash;
  unsigned int update_entry = MIB_FALSE;
  SaHpiCtrlRecDigitalT digital;
  SaHpiCtrlRecDiscreteT discrete;
  SaHpiCtrlRecAnalogT analog;
  SaHpiCtrlRecStreamT stream;
  SaHpiCtrlRecTextT text;
  SaHpiCtrlRecOemT oem;

  // Make sure they are valid.
  if (entry && ctx)
    {

      // We are subtracting SaHpiTextBufferT b/c the underlaying HPI
      // library is not zeroing out the memory for not used entries -
      // thus garbage in SaHpiTextBufferT exist,
      hash = calculate_hash_value (entry, sizeof (SaHpiCtrlRecT));

      DEBUGMSGTL ((AGENT, " Hash value: %d, in ctx: %d\n", hash, ctx->hash));

      if (ctx->hash != 0)
	{
	  // Only do the check if the hash value is something else than zero.
	  // 'zero' value is only for newly created records, and in some
	  // rare instances when the hash has rolled to zero - in which
	  // case we will just consider the worst-case scenario and update
	  // the record and not trust the hash value.
	  if (hash == ctx->hash)
	    {
	      // The same data. No need to change.
	      return AGENT_ENTRY_EXIST;
	    }
	  if (((ctx->resource_id == rpt_entry->ResourceId) &&
	      (ctx->domain_id == rpt_entry->DomainId) &&
	      (ctx->saHpiCtrlNum == entry->Num)) || 
	      (ctx->rdr_id == rdr_id)) {
		  // Time to update record.
		DEBUGMSGTL((AGENT,"Updating control entry [%d, %d, %d (rdr: %d)].\n",
					rpt_entry->DomainId,
					rpt_entry->ResourceId,
					entry->Num,
					rdr_id));
		update_entry = MIB_TRUE;
	  }
	}
      if (hash == 0)
	hash = 1;
      ctx->hash = hash;

      ctx->resource_id = rpt_entry->ResourceId;
      ctx->domain_id = rpt_entry->DomainId;
      ctx->rdr_id = rdr_id;
      ctx->saHpiCtrlNum = entry->Num;
      ctx->saHpiCtrlBool =
	(entry->Ignore == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

      /*
         entry->Type = SAHPI_CTRL_TYPE_TEXT+1;
         entry->TypeUnion.Text.MaxChars = 70;
         entry->TypeUnion.Text.MaxLines =2;
         entry->TypeUnion.Text.Language = SAHPI_LANG_DUTCH+1;
         entry->TypeUnion.Text.DataType = SAHPI_TL_TYPE_ASCII6+1;
         entry->TypeUnion.Text.Default.Text.DataLength = 28;
         strncpy(entry->TypeUnion.Text.Default.Text.Data, "Grozny smok czeka na Ciebie!", 28);

         state->StateUnion.Text.Line = 2;
         state->StateUnion.Text.Text.DataType = SAHPI_TL_TYPE_BCDPLUS+1;
         state->StateUnion.Text.Text.Language = SAHPI_LANG_DUTCH+1;
         strncpy(state->StateUnion.Text.Text.Data,"Potworny ogien jest pod Krakowem", 30);
         state->StateUnion.Text.Text.DataLength = 30;
         DEBUGMSGTL((AGENT,"[%s]\n",state->StateUnion.Text.Text.Data));
       */
      if (entry->Ignore != SAHPI_TRUE)
	{
	  // IBM-KR: Adding +1 to be compliant with the HPI MIB.
	  ctx->saHpiCtrlOutputType = entry->OutputType + 1;
	  ctx->saHpiCtrlType = entry->Type + 1;
	  // IBM-KR: End of adding +1
	  ctx->saHpiCtrlOem = entry->Oem;

	  ctx->saHpiCtrlRDR_len = rdr_entry_oid_len * sizeof (oid);
	  memcpy (ctx->saHpiCtrlRDR, rdr_entry, ctx->saHpiCtrlRDR_len);

	  // Unions
	  memset (ctx->saHpiCtrlState, 0x00, SAHPI_CTRL_STATE_MAX);
	  ctx->saHpiCtrlState_len = 0;

	  memset (ctx->saHpiCtrlAttributes, 0x00, SAHPI_CTRL_ATTR_MAX);
	  ctx->saHpiCtrlAttributes_len = 0;

	  switch (entry->Type)
	    {
	    case SAHPI_CTRL_TYPE_DIGITAL:
	      // SaHpiCtrlRecDigitalT
	      digital = entry->TypeUnion.Digital;
	      digital.Default = htonl (digital.Default);
	      ctx->saHpiCtrlAttributes_len = sizeof (SaHpiCtrlStateDigitalT);
	      memcpy (ctx->saHpiCtrlAttributes,
		      &digital.Default, ctx->saHpiCtrlState_len);
	      // SaHpiCtrlStateDigitalT
	      state->StateUnion.Digital = htonl (state->StateUnion.Digital+1);
	      ctx->saHpiCtrlState_len = sizeof (SaHpiCtrlStateDigitalT);
	      memcpy (ctx->saHpiCtrlState,
		      &state->StateUnion.Digital, ctx->saHpiCtrlState_len);
	      break;
	    case SAHPI_CTRL_TYPE_DISCRETE:
	      // SaHpiCtrlStateDiscreteT
	      discrete = entry->TypeUnion.Discrete;
	      discrete.Default = htonl (discrete.Default);
	      ctx->saHpiCtrlAttributes_len = sizeof (SaHpiCtrlStateDiscreteT);
	      memcpy (ctx->saHpiCtrlAttributes,
		      &discrete.Default, ctx->saHpiCtrlAttributes_len);
	      //SaHpiCtrlStateDiscreteT
	      state->StateUnion.Discrete = htonl (state->StateUnion.Discrete);
	      ctx->saHpiCtrlState_len = sizeof (SaHpiCtrlStateDiscreteT);
	      memcpy (ctx->saHpiCtrlState,
		      &state->StateUnion.Discrete, ctx->saHpiCtrlState_len);
	      break;
	    case SAHPI_CTRL_TYPE_ANALOG:
	      // SaHpiCtrlStateAnalogT 
	      analog = entry->TypeUnion.Analog;
	      analog.Min = htonl (analog.Min);
	      analog.Max = htonl (analog.Max);
	      analog.Default = htonl (analog.Default);
	      ctx->saHpiCtrlAttributes_len = sizeof (SaHpiCtrlRecAnalogT);
	      memcpy (ctx->saHpiCtrlAttributes,
		      &analog, ctx->saHpiCtrlAttributes_len);
	      // SaHpiCtrlStateAnalogT
	      state->StateUnion.Analog = htonl (state->StateUnion.Analog);
	      ctx->saHpiCtrlState_len = sizeof (SaHpiCtrlStateAnalogT);
	      memcpy (ctx->saHpiCtrlState,
		      &state->StateUnion.Analog, ctx->saHpiCtrlState_len);
	      break;
	    case SAHPI_CTRL_TYPE_STREAM:
	      // SaHpiCtrlRecStreamT
	      stream = entry->TypeUnion.Stream;
	      // define from SaHpi.h
	      stream.Default.StreamLength = htonl (stream.Default.StreamLength);

	      ctx->saHpiCtrlAttributes_len = sizeof (SaHpiCtrlStateStreamT);
	      memcpy (ctx->saHpiCtrlAttributes,
		      &stream.Default.Stream, ctx->saHpiCtrlAttributes_len);

	      // SaHpiCtrlStateStreamT
	      state->StateUnion.Stream.StreamLength =
		htonl (state->StateUnion.Stream.StreamLength);
	      ctx->saHpiCtrlState_len = sizeof (SaHpiCtrlStateStreamT);
	      memcpy (ctx->saHpiCtrlState,
		      &state->StateUnion.Stream, ctx->saHpiCtrlState_len);
	      break;
	    case SAHPI_CTRL_TYPE_TEXT:
	      // SaHpiCtrlRecTextT
	      text = entry->TypeUnion.Text;
	      // Max characters?
	      ctx->saHpiCtrlAttributes[0] = text.MaxChars;
	      // Max line?
	      ctx->saHpiCtrlAttributes[1] = text.MaxLines;
	      // Language/
	      ctx->saHpiCtrlAttributes[2] = text.Language+1;
	      // Text encoding
	      ctx->saHpiCtrlAttributes[3] = text.DataType+1;
	      // Line number
	      ctx->saHpiCtrlAttributes[4] = text.Default.Line;
	      // Data length
	      ctx->saHpiCtrlAttributes[5] = text.Default.Text.DataLength;
	      ctx->saHpiCtrlAttributes_len =
		(text.Default.Text.DataLength + 6 <=
		 SAHPI_CTRL_ATTR_MAX) ? text.Default.Text.
		DataLength : SAHPI_CTRL_ATTR_MAX;

	      memcpy (ctx->saHpiCtrlAttributes + 6,
		      text.Default.Text.Data, ctx->saHpiCtrlAttributes_len);

	      ctx->saHpiCtrlAttributes_len += 6;

	      //SaHpiCtrlStateTextT
	      ctx->saHpiCtrlState[0] = state->StateUnion.Text.Line;
	      ctx->saHpiCtrlState_len = sizeof (SaHpiTxtLineNumT);
	      ctx->saHpiCtrlTextType = state->StateUnion.Text.Text.DataType+1;
	      ctx->saHpiCtrlTextLanguage =
		state->StateUnion.Text.Text.Language+1;
	      ctx->saHpiCtrlText_len = state->StateUnion.Text.Text.DataLength;

	      memcpy (ctx->saHpiCtrlText,
		      state->StateUnion.Text.Text.Data,
		      ctx->saHpiCtrlText_len);
	      break;

	    case SAHPI_CTRL_TYPE_OEM:
	      oem = entry->TypeUnion.Oem;
	      oem.MId = htonl (oem.MId);
	      // manufacture id.+ OEM config data
	      ctx->saHpiCtrlAttributes_len =
		SAHPI_CTRL_OEM_CONFIG_LENGTH + sizeof (SaHpiManufacturerIdT);
	      memcpy (ctx->saHpiCtrlAttributes, &oem,
		      ctx->saHpiCtrlAttributes_len);

	      // body length (default)
	      ctx->saHpiCtrlAttributes[ctx->saHpiCtrlAttributes_len] =
		oem.Default.BodyLength;
	      ctx->saHpiCtrlAttributes_len++;

	      // default text.
	      memcpy (ctx->saHpiCtrlAttributes + ctx->saHpiCtrlAttributes_len,
		      &oem.Default.Body, oem.Default.BodyLength);

	      ctx->saHpiCtrlAttributes_len += oem.Default.BodyLength;

	      //SaHpiCtrlStateOemT
	      state->StateUnion.Oem.MId = htonl (state->StateUnion.Oem.MId);
	      ctx->saHpiCtrlState_len = sizeof (SaHpiManufacturerIdT);
	      memcpy (ctx->saHpiCtrlState,
		      &state->StateUnion.Oem.MId, ctx->saHpiCtrlState_len);
	      // We don't know the value below. Use undefined(0).
	      ctx->saHpiCtrlTextType = 0;
	      ctx->saHpiCtrlTextLanguage = 0;

	      ctx->saHpiCtrlText_len = state->StateUnion.Oem.BodyLength;
	      memcpy (ctx->saHpiCtrlText,
		      state->StateUnion.Oem.Body, ctx->saHpiCtrlText_len);
	      break;
	    }

	  if (entry->Type != state->Type)
	    {
	      DEBUGMSGTL ((AGENT, "Entry type != state type (%d=?%d)\n",
			   entry->Type, state->Type));
	      // Change 'state' to be NULL
	      ctx->saHpiCtrlState_len = 0;

	    }
	}
      if (update_entry == MIB_TRUE)
	      return AGENT_ENTRY_EXIST;
      return AGENT_NEW_ENTRY;
    }

  return AGENT_ERR_NULL_DATA;
}





/************************************************************
 * the *_row_copy routine
 */

static int
saHpiCtrlTable_row_copy (saHpiCtrlTable_context * dst,
			 saHpiCtrlTable_context * src)
{
  if (!dst || !src)
    return 1;


  if (dst->index.oids)
    free (dst->index.oids);
  if (snmp_clone_mem ((void *) &dst->index.oids, src->index.oids,
		      src->index.len * sizeof (oid)))
    {
      dst->index.oids = NULL;
      return 1;
    }
  dst->index.len = src->index.len;



  dst->saHpiCtrlNum = src->saHpiCtrlNum;

  dst->saHpiCtrlOutputType = src->saHpiCtrlOutputType;

  dst->saHpiCtrlBool = src->saHpiCtrlBool;

  dst->saHpiCtrlType = src->saHpiCtrlType;

  memcpy (dst->saHpiCtrlState, src->saHpiCtrlState, src->saHpiCtrlState_len);
  dst->saHpiCtrlState_len = src->saHpiCtrlState_len;

  memcpy (dst->saHpiCtrlAttributes, src->saHpiCtrlAttributes,
	  src->saHpiCtrlAttributes_len);
  dst->saHpiCtrlAttributes_len = src->saHpiCtrlAttributes_len;


  dst->saHpiCtrlOem = src->saHpiCtrlOem;
  dst->saHpiCtrlTextType = src->saHpiCtrlTextType;

  dst->saHpiCtrlTextLanguage = src->saHpiCtrlTextLanguage;

  memcpy (dst->saHpiCtrlText, src->saHpiCtrlText, src->saHpiCtrlText_len);
  dst->saHpiCtrlText_len = src->saHpiCtrlText_len;

  memcpy (src->saHpiCtrlRDR, dst->saHpiCtrlRDR, src->saHpiCtrlRDR_len);
  dst->saHpiCtrlRDR_len = src->saHpiCtrlRDR_len;

  dst->hash = src->hash;
  dst->resource_id = src->resource_id;
  dst->domain_id = src->domain_id;
  return 0;
}



/*
 * the *_extract_index routine
 */
int
saHpiCtrlTable_extract_index (saHpiCtrlTable_context * ctx,
			      netsnmp_index * hdr)
{
  /*
   * temporary local storage for extracting oid index
   */
  netsnmp_variable_list var_saHpiDomainID;
  netsnmp_variable_list var_saHpiResourceID;
  netsnmp_variable_list var_saHpiCtrlNum;
  int err;

  /*
   * copy index, if provided
   */
  if (hdr)
    {
      netsnmp_assert (ctx->index.oids == NULL);
      if (snmp_clone_mem ((void *) &ctx->index.oids, hdr->oids,
			  hdr->len * sizeof (oid)))
	{
	  return -1;
	}
      ctx->index.len = hdr->len;
    }

    /**
     * Create variable to hold each component of the index
     */
  memset (&var_saHpiDomainID, 0x00, sizeof (var_saHpiDomainID));
  var_saHpiDomainID.type = ASN_UNSIGNED;
  var_saHpiDomainID.next_variable = &var_saHpiResourceID;

  memset (&var_saHpiResourceID, 0x00, sizeof (var_saHpiResourceID));
  var_saHpiResourceID.type = ASN_UNSIGNED;
  var_saHpiResourceID.next_variable = &var_saHpiCtrlNum;

  memset (&var_saHpiCtrlNum, 0x00, sizeof (var_saHpiCtrlNum));
  var_saHpiCtrlNum.type = ASN_UNSIGNED;
  var_saHpiCtrlNum.next_variable = NULL;


  /*
   * parse the oid into the individual components
   */
  err = parse_oid_indexes (hdr->oids, hdr->len, &var_saHpiDomainID);
  if (err == SNMP_ERR_NOERROR)
    {
      /*
       * copy components into the context structure
       */
      ctx->saHpiCtrlNum = *var_saHpiCtrlNum.val.integer;

    }

  /*
   * parsing may have allocated memory. free it.
   */
  snmp_reset_var_buffers (&var_saHpiDomainID);

  return err;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int
saHpiCtrlTable_can_delete (saHpiCtrlTable_context * undo_ctx,
			   saHpiCtrlTable_context * row_ctx,
			   netsnmp_request_group * rg)
{


  return 1;
}


/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiCtrlTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiCtrlTable_context *
saHpiCtrlTable_create_row (netsnmp_index * hdr)
{
  saHpiCtrlTable_context *ctx = SNMP_MALLOC_TYPEDEF (saHpiCtrlTable_context);
  if (!ctx)
    return NULL;

  if (saHpiCtrlTable_extract_index (ctx, hdr))
    {
      free (ctx->index.oids);
      free (ctx);
      return NULL;
    }

  ctx->saHpiCtrlState_len = 0;
  ctx->saHpiCtrlAttributes_len = 0;
  ctx->saHpiCtrlTextType = 0;
  ctx->saHpiCtrlTextLanguage = 0;
  ctx->saHpiCtrlText_len = 0;
  ctx->hash = 0;
  return ctx;
}


/************************************************************
 * the *_duplicate row routine
 */
saHpiCtrlTable_context *
saHpiCtrlTable_duplicate_row (saHpiCtrlTable_context * row_ctx)
{
  saHpiCtrlTable_context *dup;

  if (!row_ctx)
    return NULL;

  dup = SNMP_MALLOC_TYPEDEF (saHpiCtrlTable_context);
  if (!dup)
    return NULL;

  if (saHpiCtrlTable_row_copy (dup, row_ctx))
    {
      free (dup);
      dup = NULL;
    }

  return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index *
saHpiCtrlTable_delete_row (saHpiCtrlTable_context * ctx)
{
  /*
   * netsnmp_mutex_destroy(ctx->lock); 
   */

  if (ctx->index.oids)
    free (ctx->index.oids);


  free (ctx);

  return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void
saHpiCtrlTable_set_reserve1 (netsnmp_request_group * rg)
{
  saHpiCtrlTable_context *row_ctx =
    (saHpiCtrlTable_context *) rg->existing_row;

  netsnmp_variable_list *var;
  netsnmp_request_group_item *current;

  int rc;

  DEBUGMSGTL ((AGENT, "saHpiCtrlTable_set_reserve1: Entry\n"));

  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;
      rc = SNMP_ERR_NOERROR;

      switch (current->tri->colnum)
	{


	case COLUMN_SAHPICTRLNUM:
	case COLUMN_SAHPICTRLOUTPUTTYPE:
	case COLUMN_SAHPICTRLBOOL:
	case COLUMN_SAHPICTRLOEM:
	case COLUMN_SAHPICTRLRDR:
	  rc = SNMP_ERR_NOTWRITABLE;
	  break;
	case COLUMN_SAHPICTRLSTATE:
	  /** OCTETSTR = ASN_OCTET_STR */
	  if (var->type != ASN_OCTET_STR)
	    {
	      rc = SNMP_ERR_WRONGTYPE;
	    }
	  else if (var->val_len > SAHPI_CTRL_STATE_MAX)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  else if (var->val_len < 0)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  break;

	case COLUMN_SAHPICTRLTEXTTYPE:
	    /** INTEGER = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiCtrlTextType));
	  break;

	case COLUMN_SAHPICTRLTEXTLANGUAGE:
	    /** INTEGER = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiCtrlTextLanguage));
	  break;

	case COLUMN_SAHPICTRLTEXT:
	  /** OCTETSTR = ASN_OCTET_STR */
	  if (var->type != ASN_OCTET_STR)
	    {
	      rc = SNMP_ERR_WRONGTYPE;
	    }
	  else if (var->val_len > SAHPI_TEXT_MAX)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  else if (var->val_len < 0)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  break;
	default:
		/** We shouldn't get here */
	  rc = SNMP_ERR_GENERR;
	  snmp_log (LOG_ERR, "unknown column in "
		    "saHpiCtrlTable_set_reserve1\n");
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
      rg->status = SNMP_MAX (rg->status, current->ri->status);
    }
  for (current = rg->list; current; current = current->next)
    {

      // The nice thing about this API is that _row_copy() is called
      // for this row - if the API has matched the index with an
      // already existing entry. We check the 'hash' value. If its
      // 0 the API couldn't find the right context.

      if (row_ctx->hash == 0)
	{
	  rc = SNMP_ERR_NOSUCHNAME;
	  netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
	}
    }
  DEBUGMSGTL ((AGENT, "saHpiCtrlTable_set_reserve1: Exit\n"));
}

void
saHpiCtrlTable_set_reserve2 (netsnmp_request_group * rg)
{
  saHpiCtrlTable_context *row_ctx =
    (saHpiCtrlTable_context *) rg->existing_row;

  netsnmp_request_group_item *current;
  netsnmp_variable_list *var;
  int rc;

  rg->rg_void = rg->list->ri;
  DEBUGMSGTL ((AGENT, "saHpiCtrlTable_set_reserve2: Entry\n"));

  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;
      rc = SNMP_ERR_NOERROR;

      switch (current->tri->colnum)
	{

	case COLUMN_SAHPICTRLSTATE:
	  // More detail check
	  switch (row_ctx->saHpiCtrlType - 1)
	    {
	    case SAHPI_CTRL_TYPE_DIGITAL:
	      if (var->val_len != sizeof (SaHpiCtrlStateDigitalT))
		rc = SNMP_ERR_WRONGLENGTH;
	      break;
	    case SAHPI_CTRL_TYPE_DISCRETE:
	      if (var->val_len != sizeof (SaHpiCtrlStateDiscreteT))
		rc = SNMP_ERR_WRONGLENGTH;
	      break;
	    case SAHPI_CTRL_TYPE_ANALOG:
	      if (var->val_len != sizeof (SaHpiCtrlStateAnalogT))
		rc = SNMP_ERR_WRONGLENGTH;
	      break;
	    case SAHPI_CTRL_TYPE_STREAM:
	      if (var->val_len != sizeof (SaHpiCtrlStateStreamT))
		rc = SNMP_ERR_WRONGLENGTH;
	      break;
	    case SAHPI_CTRL_TYPE_TEXT:
	      if (var->val_len != sizeof (SaHpiTxtLineNumT))
		rc = SNMP_ERR_WRONGLENGTH;
	      break;
	    case SAHPI_CTRL_TYPE_OEM:
	      if (var->val_len != sizeof (SaHpiManufacturerIdT))
		rc = SNMP_ERR_WRONGLENGTH;
	      break;
	    default:
	      // IBM-KR: Perhaps another error code. 
	      rc = SNMP_ERR_BADVALUE;
	    }
	  break;

	case COLUMN_SAHPICTRLTEXTTYPE:
	  if (((*var->val.integer < 1) || (*var->val.integer > 4)))
	    {
	      rc = SNMP_ERR_BADVALUE;
	    }
	  switch (row_ctx->saHpiCtrlType - 1)
	    {
	    case SAHPI_CTRL_TYPE_TEXT:
	      break;
	    case SAHPI_CTRL_TYPE_OEM:
	      break;
	    default:
	      // For all other types:
	      rc = SNMP_ERR_NOTWRITABLE;
	      break;
	    }
	  break;

	case COLUMN_SAHPICTRLTEXTLANGUAGE:

	  if (((*var->val.integer < 1) || (*var->val.integer > 137)))
	    {
	      rc = SNMP_ERR_BADVALUE;
	    }
	  switch (row_ctx->saHpiCtrlType - 1)
	    {
	    case SAHPI_CTRL_TYPE_TEXT:
	      break;
	    case SAHPI_CTRL_TYPE_OEM:
	      break;
	    default:
	      // For all other types:
	      rc = SNMP_ERR_NOTWRITABLE;
	      break;
	    }
	  break;

	case COLUMN_SAHPICTRLTEXT:
	  // Write can only happend if type is OEM or Text.
	  switch (row_ctx->saHpiCtrlType - 1)
	    {
	    case SAHPI_CTRL_TYPE_TEXT:
	      break;
	    case SAHPI_CTRL_TYPE_OEM:
	      break;
	    default:
	      // For all other types:
	      rc = SNMP_ERR_NOTWRITABLE;
	      break;
	    }
	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	  break;
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
    }
  DEBUGMSGTL ((AGENT, "saHpiCtrlTable_set_reserve2: Exit\n"));
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void
saHpiCtrlTable_set_action (netsnmp_request_group * rg)
{
  netsnmp_variable_list *var;
  saHpiCtrlTable_context *row_ctx =
    (saHpiCtrlTable_context *) rg->existing_row;

  netsnmp_request_group_item *current;

  int rc = SNMP_ERR_NOERROR;

  DEBUGMSGTL ((AGENT, "saHpiCtrlTable_set_action: Entry\n"));
  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;

      switch (current->tri->colnum)
	{

	case COLUMN_SAHPICTRLSTATE:
	  memcpy (row_ctx->saHpiCtrlState, var->val.string, var->val_len);
	  row_ctx->saHpiCtrlState_len = var->val_len;

	  if (row_ctx->saHpiCtrlType - 1 == SAHPI_CTRL_TYPE_TEXT)
	    {
	      // Want to read a new line.
	      if (read_textline (row_ctx) != AGENT_ERR_NOERROR)
		{
		  // Would be nice to have : "Wrong line error"
		  rc = SNMP_ERR_BADVALUE;
		}
	    }
	  else
	    {
	      if (set_ctrl_state (row_ctx) != AGENT_ERR_NOERROR)
		rc = SNMP_ERR_INCONSISTENTVALUE;
	    }
	  break;

	case COLUMN_SAHPICTRLTEXTTYPE:
	    /** INTEGER = ASN_INTEGER */
	  row_ctx->saHpiCtrlTextType = *var->val.integer;
	  break;

	case COLUMN_SAHPICTRLTEXTLANGUAGE:
	    /** INTEGER = ASN_INTEGER */
	  row_ctx->saHpiCtrlTextLanguage = *var->val.integer;
	  break;

	case COLUMN_SAHPICTRLTEXT:
	    /** OCTETSTR = ASN_OCTET_STR */
	  memcpy (row_ctx->saHpiCtrlText, var->val.string, var->val_len);
	  row_ctx->saHpiCtrlText_len = var->val_len;

	  if (set_ctrl_state (row_ctx) != AGENT_ERR_NOERROR)
	    rc = SNMP_ERR_INCONSISTENTVALUE;

	  break;

	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
    }
  DEBUGMSGTL ((AGENT, "saHpiCtrlTable_set_action: Exit\n"));
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void
saHpiCtrlTable_set_commit (netsnmp_request_group * rg)
{

}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void
saHpiCtrlTable_set_free (netsnmp_request_group * rg)
{

}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void
saHpiCtrlTable_set_undo (netsnmp_request_group * rg)
{

}




/************************************************************
 *
 * Initialize the saHpiCtrlTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiCtrlTable (void)
{
  netsnmp_table_registration_info *table_info;

  if (my_handler)
    {
      snmp_log (LOG_ERR,
		"initialize_table_saHpiCtrlTable_handler called again\n");
      return;
    }

  memset (&cb, 0x00, sizeof (cb));

    /** create the table structure itself */
  table_info = SNMP_MALLOC_TYPEDEF (netsnmp_table_registration_info);

  /*
   * if your table is read only, it's easiest to change the
   * HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY 
   */
  my_handler = netsnmp_create_handler_registration ("saHpiCtrlTable",
						    netsnmp_table_array_helper_handler,
						    saHpiCtrlTable_oid,
						    saHpiCtrlTable_oid_len,
						    HANDLER_CAN_RWRITE);

  if (!my_handler || !table_info)
    {
      snmp_log (LOG_ERR, "malloc failed in "
		"initialize_table_saHpiCtrlTable_handler\n");
      return;	/** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */


  /*
   * internal indexes
   */
	/** index: saHpiCtrlNum */
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);

  table_info->min_column = saHpiCtrlTable_COL_MIN;
  table_info->max_column = saHpiCtrlTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
  cb.get_value = saHpiCtrlTable_get_value;
  cb.container = netsnmp_container_find ("saHpiCtrlTable_primary:"
					 "saHpiCtrlTable:" "table_container");


  cb.create_row = (UserRowMethod *) saHpiCtrlTable_create_row;

  cb.duplicate_row = (UserRowMethod *) saHpiCtrlTable_duplicate_row;
  cb.delete_row = (UserRowMethod *) saHpiCtrlTable_delete_row;
  cb.row_copy = (Netsnmp_User_Row_Operation *) saHpiCtrlTable_row_copy;

  cb.can_delete = (Netsnmp_User_Row_Action *) saHpiCtrlTable_can_delete;

  cb.set_reserve1 = saHpiCtrlTable_set_reserve1;
  cb.set_reserve2 = saHpiCtrlTable_set_reserve2;
  cb.set_action = saHpiCtrlTable_set_action;
  cb.set_commit = saHpiCtrlTable_set_commit;
  cb.set_free = saHpiCtrlTable_set_free;
  cb.set_undo = saHpiCtrlTable_set_undo;


  DEBUGMSGTL (("initialize_table_saHpiCtrlTable",
	       "Registering table saHpiCtrlTable " "as a table array\n"));
  netsnmp_table_container_register (my_handler, table_info, &cb,
				    cb.container, 1);



  netsnmp_register_read_only_counter32_instance ("control_count",
						 saHpiCtrlCount_oid,
						 OID_LENGTH
						 (saHpiCtrlCount_oid),
						 &control_count, NULL);
}

/************************************************************
 * saHpiCtrlTable_get_value
 */
int
saHpiCtrlTable_get_value (netsnmp_request_info * request,
			  netsnmp_index * item,
			  netsnmp_table_request_info * table_info)
{
  netsnmp_variable_list *var = request->requestvb;
  saHpiCtrlTable_context *context = (saHpiCtrlTable_context *) item;

  switch (table_info->colnum)
    {

    case COLUMN_SAHPICTRLNUM:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->saHpiCtrlNum,
				sizeof (context->saHpiCtrlNum));
      break;

    case COLUMN_SAHPICTRLOUTPUTTYPE:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiCtrlOutputType,
				sizeof (context->saHpiCtrlOutputType));
      break;

    case COLUMN_SAHPICTRLBOOL:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiCtrlBool,
				sizeof (context->saHpiCtrlBool));
      break;

    case COLUMN_SAHPICTRLTYPE:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiCtrlType,
				sizeof (context->saHpiCtrlType));
      break;

    case COLUMN_SAHPICTRLSTATE:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->saHpiCtrlState,
				context->saHpiCtrlState_len);
      break;

    case COLUMN_SAHPICTRLATTRIBUTES:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->saHpiCtrlAttributes,
				context->saHpiCtrlAttributes_len);
      break;

    case COLUMN_SAHPICTRLOEM:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->saHpiCtrlOem,
				sizeof (context->saHpiCtrlOem));
      break;

    case COLUMN_SAHPICTRLTEXTTYPE:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiCtrlTextType,
				sizeof (context->saHpiCtrlTextType));
      break;

    case COLUMN_SAHPICTRLTEXTLANGUAGE:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiCtrlTextLanguage,
				sizeof (context->saHpiCtrlTextLanguage));
      break;

    case COLUMN_SAHPICTRLTEXT:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->saHpiCtrlText,
				context->saHpiCtrlText_len);
      break;

    case COLUMN_SAHPICTRLRDR:
	    /** RowPointer = ASN_OBJECT_ID */
      snmp_set_var_typed_value (var, ASN_OBJECT_ID,
				(char *) &context->saHpiCtrlRDR,
				context->saHpiCtrlRDR_len);
      break;

    default:
	    /** We shouldn't get here */
      snmp_log (LOG_ERR, "unknown column in " "saHpiCtrlTable_get_value\n");
      return SNMP_ERR_GENERR;
    }
  return SNMP_ERR_NOERROR;
}
